import inspect
import uuid
from typing import List, Tuple, Any, Callable, Awaitable
from uuid import uuid4
from ..storage.base import StorageManager
from ..core.types.client import PaymentPayload
from ..core.types.facilitator import VerifyRequest, X402_VERSION, FacilitatorConfig, SettleRequest, SettleResponse
from ..core.types.requirements import PaymentRequirements
from ..core.types.setup import Y402Setup
from ..facilitator_client.httpx import FacilitatorClient
from ..lifecycle.utils import create_settled_payment


async def _maybe_await(result: Any) -> Any:
    if inspect.isawaitable(result):
        result = await result
    return result


async def process_payment(
    # The identity of the payment.
    resource: str, tags: List[str], reference: str, endpoint: Callable[[uuid4], tuple | Awaitable[tuple]],
    # User payment selection.
    payment: PaymentPayload, matched_requirements: PaymentRequirements,
    # External components.
    setup: Y402Setup, facilitator_config: FacilitatorConfig,
    # Storage.
    storage_manager: StorageManager, storage_collection: str,
    # Webhook-related data.
    webhook_name: str,
    # Tunings.
    request_timeout: int = 15
) -> Tuple[uuid.UUID, SettleResponse, Any]:
    """
    Processes a given payment.

    Args:
        resource: The (PUBLIC) resource URL.
        tags: The tags that apply.
        reference: The reference that applies.
        endpoint: The (wrapped) endpoint to invoke.
        payment: The user-submitted payment.
        matched_requirements: The matched requirements.
        setup: An existing Y402 setup.
        facilitator_config: The facilitator config to use to create a client.
        storage_manager: The storage manager for payments.
        storage_collection: The collection to store the payment into.
        webhook_name: The name of the webhook.
        request_timeout: The timeout for requests.

    Returns:
        The processed UUID for this payment. Also, the settle response, if any.
        Finally, the underlying response from the endpoint. If the response is
        not a [200..399] response, settle response will be null.
    """

    # 1. Prepare all the data for a settled payment. Do this in advance
    #    so no failures are spotted later, on settling, on this topic.
    payer = payment.payload.authorization.from_
    network = payment.network
    token = matched_requirements.asset
    value = payment.payload.authorization.value
    chain_id, code, name, price_label = setup.get_payment_data(network, token, value)
    payment_id = uuid4()
    settled_payment = create_settled_payment(
        payment_id, resource, tags, reference,
        payer, chain_id, token, value, payment.payload.authorization.to,
        code, name, price_label
    )

    # 2. Create the facilitator config.
    facilitator_client = FacilitatorClient(facilitator_config)

    # 3. Perform the verification.
    await facilitator_client.verify(VerifyRequest(
        x402Version=X402_VERSION,
        paymentPayload=payment,
        paymentRequirements=matched_requirements,
        timeout=request_timeout
    ))

    # 4. Store the verified payment. Done so, if the settlement
    #    fails and the client complains that the authorization
    #    was consumed, then the data is present for any claim.
    #    Abusing this system would involve a decent amount of
    #    gas consumption from the client, so very often this
    #    will be legit.
    await _maybe_await(storage_manager.allocate(storage_collection, payment_id, payment, matched_requirements,
                                                settled_payment, webhook_name))

    # 5. Execute the underlying endpoint.
    response, success = _maybe_await(endpoint(payment_id))

    if success:
        # 6. Settle the payment. By this point, the payment is consumed
        #    and the corresponding record is marked as such.
        settle_response = await facilitator_client.settle(SettleRequest(
            x402Version=X402_VERSION,
            paymentPayload=payment,
            paymentRequirements=matched_requirements
        ), timeout=request_timeout)

        # 7. Here, the response was successful in terms of HTTP (200)
        #    and the settling was successful as well.
        #
        #    If the setting were not to be successful, then it would
        #    not execute this point (committing the settlement).
        if settle_response.success:
            await _maybe_await(storage_manager.settle(
                storage_collection, payment_id, settle_response.transaction
            ))
    else:
        settle_response = None
        await _maybe_await(storage_manager.abort(storage_collection, payment_id))
    return payment_id, settle_response, response



from typing import Optional, Literal
from ..core.types.facilitator import VerifyResponse, VerifyRequest, SettleRequest, SettleResponse, \
    FacilitatorConfig
from .errors import VerifyFacilitatorInvalidError, VerifyBadResponse, SettleBadResponse, SettleFacilitatorFailedError
from .utils import make_headers


class FacilitatorClient:
    """
    This class defines the base methods for facilitator clients.
    """

    def __init__(self, config: Optional[FacilitatorConfig] = None):
        self._config = config or FacilitatorConfig()

    def _make_headers(self, endpoint: Literal['verify', 'settle']) -> dict:
        """
        Makes the headers for the chosen endpoint and out of the
        settings given for the header.

        Args:
            endpoint: 'verify' or 'settle'.

        Returns:
            The dictionary of headers.
        """

        return make_headers(self._config.headers or {}, endpoint)

    def _check_verify_status(self, status_code: int, content: bytes, content_type: str):
        """
        Requires the status_code of verify to be 2xx.

        Args:
            status_code: The status code.
        """

        if status_code not in range(200, 300):
            raise VerifyBadResponse(status_code, content, content_type)

    def _parse_verify_obj(self, obj: dict) -> VerifyResponse:
        """
        Parses a verify object (dict) into a model. It also
        fails if the verification is not valid.

        Args:
            obj: The dictionary to parse.

        Returns:
            The parsed object.
        """

        obj = VerifyResponse(**obj)
        if not obj.is_valid:
            raise VerifyFacilitatorInvalidError(obj)
        return obj

    def verify(self, request: VerifyRequest, timeout: int = 10) -> VerifyResponse:
        """
        Performs a /verify POST call with the given data.

        Args:
            request: The current request.
            timeout: The timeout.

        Returns:
            The verify response.
        """

        raise NotImplementedError

    def _check_settle_status(self, status_code: int, content: bytes, content_type: str):
        """
        Requires the status_code of settle to be 2xx.

        Args:
            status_code: The status code.
        """

        if status_code not in range(200, 300):
            raise SettleBadResponse(status_code, content, content_type)

    def _parse_settle_obj(self, obj: dict) -> SettleResponse:
        """
        Parses a settle object (dict) into a model. It also
        fails if the settling has failed.

        Args:
            obj: The dictionary to parse.

        Returns:
            The parsed object.
        """

        obj = SettleResponse(**obj)
        if not obj.success:
            raise SettleFacilitatorFailedError(obj)
        return obj

    def settle(self, request: SettleRequest, timeout: int = 10) -> SettleResponse:
        """
        Performs a /settle POST call with the given data.

        Args:
            request: The current request.
            timeout: The timeout.

        Returns:
            The settle response.
        """

        raise NotImplementedError



from typing import Optional, Dict, Literal, Any
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class HTTPInputSchema(BaseModel):
    """
    Schema for HTTP request input, excluding spec
    and method which are handled by the middleware.
    """

    query_params: Optional[Dict[str, str]] = None
    body_type: Optional[
        Literal["json", "form-data", "multipart-form-data", "text", "binary"]
    ] = None
    body_fields: Optional[Dict[str, Any]] = None
    header_fields: Optional[Dict[str, Any]] = None

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )



from typing import Tuple, Optional, List, Dict
from decimal import Decimal
from pydantic import BaseModel
from .errors import MisconfigurationError
from .default_data import KNOWN_NETWORKS_AND_TOKENS


class Y402Setup(BaseModel):
    """
    A setup tells in a set of internal mappings:

    - The available networks.
    - The tokens in those networks.
    """

    _networks: Optional[dict] = None
    _default_tokens: Optional[dict] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._networks = self._networks or {}
        self._default_tokens = self._default_tokens or {}

    def add_network(self, network: str, chain_id: int = 0):
        """
        Adds a network spec to the current setup. This
        is done by telling which network and also the
        chain id, but the chain id can be omitted if
        the network is known (in fact: it will be not
        considered, but replaced / reset, if it's for
        a known network name).

        Args:
            network: The network id.
            chain_id: The chain id.
        """

        network = network.strip().lower()
        if network in self._networks:
            raise MisconfigurationError(f"This network is already set up: {network}")
        if network in KNOWN_NETWORKS_AND_TOKENS and chain_id < 1:
            chain_id = KNOWN_NETWORKS_AND_TOKENS[network]["chain_id"]
        if chain_id < 1:
            raise MisconfigurationError(f"Invalid chain id for network key: {network}")
        self._networks[network] = {
            "chain_id": chain_id,
            "tokens": {},
            "tokens_by_address": {},
            "default_token": None
        }

    def add_token(self, network: str, code: str,
                  # The name.
                  name: str = "",
                  # The address and the version.
                  address: str = "", version: str = "",
                  # The decimals and the symbol.
                  decimals: int = 0, symbol: str = "",
                  # Whether it's default or not.
                  default_for_symbol: bool = False):
        """
        Adds a token contract to a specific place.

        Args:
            network: The network name for which the token will be added.
            code: An internal name for this contract, like "usdc". Given
                  an existing network and a known code, adding it will
                  just use the names of the known data, optionally with
                  some overrides (e.g. the version).
            name: A name for this contract. It may be chosen, e.g. "USDC".
            address: The address of the token.
            version: The EIP-712 version of the token. If the token is
                     an EIP-3009 token it should have a version. Try
                     "1" by default is not, or a value that is sensible
                     for the contract itself when trying.
            decimals: The amount of decimals of the token. It MUST match
                      the value returned at .decimals() in the token.
            symbol: A symbol for this contract, like $ or €. It is optional.
            default_for_symbol: Whether the current token will be resolved
                                to be the one to use when its symbol is
                                used in a price with that symbol (or no
                                symbol, perhaps).
        """

        network = network.strip().lower()
        code = code.strip().lower()
        if network not in self._networks:
            raise MisconfigurationError(f"This network is not yet set up: {network}")
        if not code:
            raise MisconfigurationError(f"Use a valid code for the token")
        if code in self._networks[network]["tokens"]:
            raise MisconfigurationError(f"This token is already set up in network {network}: {code}")
        if symbol in " 0123456789." or len(symbol) > 1:
            raise MisconfigurationError(f"Invalidd symbol: {symbol}")

        if code in KNOWN_NETWORKS_AND_TOKENS[network]["tokens"]:
            known_token = KNOWN_NETWORKS_AND_TOKENS[network]["tokens"][code]
            decimals = decimals or known_token["decimals"]
            version = version or known_token["eip712Version"]
            symbol = symbol or known_token["symbol"]
            address = address or known_token["address"]
            name = name or known_token["name"]

        name = name.strip() or code.upper()
        if not address or not version or decimals < 0 or not name:
            raise MisconfigurationError(f"One or more token arguments are missing")
        if address in self._networks[network]["tokens_by_address"]:
            raise MisconfigurationError(f"This token's address is already setup in network {network}: {address}")

        self._networks[network]["tokens"][code] = {
            "name": name, "symbol": symbol, "address": address,
            "version": version, "decimals": decimals
        }
        self._networks[network]["tokens_by_address"][address] = code
        if default_for_symbol:
            self._default_tokens.setdefault(network, {})[symbol] = code

    def _check_network_and_code(self, network: str, code: str) -> Tuple[str, str]:
        """
        Checks the chosen network and token code to exist.

        Args:
            network: The network name for which the token will be added.
            code: An internal name for an existing token contract in the
                  network in this setup.
        """

        network = network.strip().lower()
        code = code.strip().lower()
        if network not in self._networks:
            raise MisconfigurationError(f"This network is not yet set up: {network}")
        if not code:
            raise MisconfigurationError(f"Use a valid code for the token")
        if code not in self._networks[network]["tokens"]:
            raise MisconfigurationError(f"This token is not yet set up in network {network}: {code}")

        return network, code

    def set_default_for_symbol_token(self, network: str, code: str):
        """
        Given a network code and a token code, it ensures the token
        becomes the default one for its symbol (even if for some
        reason no symbol were to be defined in a token).

        Args:
            network: The network name for which the token will be added.
            code: An internal name for an existing token contract in the
                  network in this setup.
        """

        network, code = self._check_network_and_code(network, code)
        token = self._networks[network]["tokens"][code]
        symbol = token["symbol"]
        self._default_tokens.setdefault(network, {})[symbol] = code

    def set_default_token(self, network: str, code: str):
        """
        Given a network code and a token code, it ensures the token
        becomes the default one, this time not for its symbol but
        instead for when an integer value is used.

        Args:
            network: The network name for which the token will be defaulted.
            code: An internal name for an existing token contract in the
                  network in this setup.
        """

        network, code = self._check_network_and_code(network, code)
        self._networks[network]["default_token"] = code

    def get_default_token(self, network: str) -> Optional[str]:
        """
        Gets the default token of a network

        Args:
            network: The network name for which the token will be defaulted.

        Returns:
            The default token code of that network.
        """

        if network not in self._networks:
            return None
        return self._networks[network]["default_token"]

    def _get_price_label(self, value: str, decimals: int, symbol: str):
        """
        Builds a price label for an amount of a given currency.

        Args:
            value: A decimal representation of the amount.
            decimals: The amount of digits that are decimal places.
            symbol: The currency symbol.

        Returns:
            A textual representation.
        """

        d = Decimal(value) / (10 ** decimals)
        return f"{symbol}{d}"

    def list_tokens(self, network: str) -> List[str]:
        """
        Returns the list of registered token (codes) for a network.

        Args:
            network: The network name for which the tokens will be listed.

        Returns:
            The list of registered tokens.
        """

        network = network.strip().lower()
        if network not in self._networks:
            raise MisconfigurationError(f"This network is not yet set up: {network}")
        return list(self._networks[network]["tokens"].keys())

    def get_token_metadata(self, network: str, code: str) -> Tuple[str, str, str, str, int]:
        """
        Returns the metadata associated to a token.

        Args:
            network: The network name for which the token will be defaulted.
            code: An internal name for an existing token contract in the
                  network in this setup.

        Returns:
            The associated metadata.
        """

        network, code = self._check_network_and_code(network, code)
        token = self._networks[network]["tokens"][code]
        return token["name"], token["symbol"], token["address"], token["version"], token["decimals"]

    def get_payment_data(self, network: str, token: str, value: str) -> tuple[int, str, str, str]:
        """
        Returns data associated to a specific token payment.

        Args:
            network: The name of the network.
            token: The address of the token contract.
            value: A decimal representation of the amount.

        Returns:
            A tuple (chain_id, code, name, price_label).
        """

        network = network.strip().lower()
        if network not in self._networks:
            raise MisconfigurationError(f"This network is not yet set up: {network}")
        if token not in self._networks[network]["tokens_by_address"]:
            raise MisconfigurationError(f"Use a valid code for the token")
        code = self._networks[network]["tokens_by_address"][token]
        token_data = self._networks[network]["tokens"][code]
        decimals = token_data["decimals"]
        name = token_data["name"]
        chain_id = self._networks[network]["chain_id"]
        symbol = token_data["symbol"]
        return chain_id, code, name, self._get_price_label(value, decimals, symbol)

    def get_chain_id(self, network: str) -> int:
        """
        Given a network, returns its chain id.

        Args:
            network: The name of the network.

        Returns:
            An integer being the chain id.
        """

        try:
            return self._networks[network]["chain_id"]
        except:
            raise MisconfigurationError(f"This network is not set up: {network}")

    def get_chain_ids_mapping(self) -> Dict[str, int]:
        """
        Returns the available mapping of name => chain_id.

        Returns:
            A dictionary mapping name => chain_id from this setup.
        """

        return {key: value["chain_id"] for key, value in self._networks.items()}

    def parse_price_label(self, network: str, label: str) -> tuple[str, str]:
        """
        Given a price label, it tries to parse it.

        Args:
            network: The name of the network.
            label: The label to parse.

        Returns:
            The parsed token price, as (asset code, amount).
        """

        label = label.strip()
        if not label:
            return "", "0"

        # 1. Parse the symbol and get the token.
        if label[0] not in "0123456789.":
            symbol, price = label[0], label[1:]
        else:
            symbol, price = "", label
        if symbol not in self._default_tokens[network]:
            raise MisconfigurationError(f"The symbol '{symbol}' is not default-registered in network: {network}")
        code = self._default_tokens[network][symbol]
        token_data = self._networks[network]["tokens"][code]
        decimals = token_data["decimals"]

        # 2. Parse the price and multiply by decimals to get the amount.
        try:
            d = Decimal(price)
            if d < 0:
                raise ValueError("Invalid Price")
            amount = str(int(d * 10 ** decimals))
        except:
            raise MisconfigurationError(f"The price '{price} is not a valid amount")

        # 3. Return the token code and the amount.
        return code, amount

    def __or__(self, other: 'Y402Setup'):
        """
        Merges two definitions (as a new definition) when doing this. The
        definitions in the first operand take precedence when defining
        networks and tokens, but the definitions in the second operand
        take precedence when defining defaults.
        :param other: The other setup to merge.
        :return: The new, merged, definition.
        """

        if not isinstance(other, Y402Setup):
            raise TypeError(f"Can only merge Y402Setup (not \"{type(other).__name__}\") to Y402Setup")

        merged = Y402Setup()
        for obj in [self, other]:
            for network, values in obj._networks.items():
                # First, add the network.
                chain_id = values["chain_id"]
                tokens = values["tokens"]
                try:
                    merged.add_network(network, chain_id)
                except:
                    pass

                # Then, add the tokens.
                for code, token_data in tokens.items():
                    name = token_data["name"]
                    symbol = token_data["symbol"]
                    address = token_data["address"]
                    version = token_data["version"]
                    decimals = token_data["decimals"]
                    try:
                        obj.add_token(network, code, name, address, version, decimals, symbol)
                    except:
                        pass

                # Finally, add the defaults. The defaults
                # in the other setup will take precedence.
                for symbol, code in obj._default_tokens.get(network, {}):
                    merged.set_default_for_symbol_token(network, code)

        return merged



from typing import Optional, Any, Union, Literal
import re
from pydantic import BaseModel, Field, ConfigDict, field_validator
from pydantic.alias_generators import to_camel
from .eip712 import EIP712Domain


class TokenAsset(BaseModel):
    """
    Represents token asset information including EIP-712
    domain data.
    """

    address: str
    decimals: int
    eip712: EIP712Domain

    @field_validator("decimals")
    def validate_decimals(cls, v):
        if v < 0 or v > 255:
            raise ValueError("decimals must be between 0 and 255")
        return v


class TokenAmount(BaseModel):
    """
    Represents an amount of tokens in atomic units with asset
    information.
    """

    amount: str
    asset: TokenAsset

    @field_validator("amount")
    def validate_amount(cls, v):
        try:
            int(v)
        except ValueError:
            raise ValueError("amount must be an integer encoded as a string")
        return v


Price = Union[str, int, TokenAmount]


class RequirePaymentDetails(BaseModel):
    """
    This is a payment requirement specification that applies
    for an endpoint in particular.
    """

    scheme: Literal["exact"] = Field(description="The scheme", default="exact")
    network: str = Field(
        description="The human name of the network (e.g. ethereum, ethereum-sepolia, "
                    "base, base-sepolia, avalanche, avalanche-fuji)"
    )
    price: Price = Field(
        description="The price, which can be any x402-supported price type"
    )
    pay_to_address: str = Field(
        description="The address to pay to. It must be a valid address"
    )

    @field_validator("pay_to_address")
    def validate_pay_to_address(cls, v):
        if not re.match(r"^0x[a-fA-F0-9]{40}$", v):
            raise ValueError("pay_to_address must be a 0x-prefixed 40-hex digits string")
        return v

    @field_validator("price")
    def validate_price(cls, v):
        has_allowed_type = True

        if not isinstance(v, (str, int, TokenAmount)):
            has_allowed_type = False
        if isinstance(v, str):
            v = v.strip()

            if v[0] not in '0123456789.':
                v_ = v[1:]
            else:
                v_ = v
            try:
                if float(v_) < 0:
                    raise Exception()
            except:
                has_allowed_type = False
        elif isinstance(v, int):
            if v < 0:
                has_allowed_type = False
        elif isinstance(v, TokenAmount):
            try:
                if int(v.amount) < 0:
                    raise Exception()
            except:
                has_allowed_type = False

        if not has_allowed_type:
            raise ValueError(f"The price must be a valid positive numeric str (can be $-prefixed), "
                             f"a positive int, or a positive TokenAmount object")

        return v


class PaymentRequirements(BaseModel):
    """
    This is the details of a payment requirement that both
    server and client agree on.
    """

    scheme: str = Field(
        default='exact',
        description="The payment payload scheme, e.g. 'exact'"
    )
    network: str = Field(
        description="The payment network by name (e.g. 'base', 'base-sepolia')"
    )
    max_amount_required: str = Field(
        description="The decimal representation of the required amount, in "
                    "terms of the minimal units"
    )
    resource: str = Field(
        description="The associated canonical resource URL or request URL for "
                    "this payment"
    )
    description: str = Field(
        description="The description of this service / the purpose of the payment"
    )
    mime_type: str = Field(
        default="application/json",
        description="The associated MIME type for this request"
    )
    output_schema: Optional[Any] = Field(
        default=None,
        description="The expected schema for this endpoint. Serves as documentation"
    )
    pay_to: str = Field(
        description="The address the payment must be authorized to"
    )
    max_timeout_seconds: int = Field(
        description="The maximum time the client is allowed to take before paying"
    )
    asset: str = Field(
        description="The 0x-prefixed ethereum address of the asset chosen "
                    "to make the payment in"
    )
    extra: Optional[dict[str, Any]] = Field(
        description="Extra data (typically used to state the EIP-712 domain"
    )

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    @field_validator("max_amount_required")
    def validate_max_amount_required(cls, v):
        try:
            int(v)
        except ValueError:
            raise ValueError(
                "max_amount_required must be an integer encoded as a string"
            )
        return v


class FinalRequiredPaymentDetails(BaseModel):
    """
    This is a final payment requirement specification
    that applies for an endpoint in particular.
    """

    scheme: str = Field(description="The scheme", default="exact")
    network: str = Field(
        description="The human name of the network (e.g. ethereum, ethereum-sepolia, "
                    "base, base-sepolia, avalanche, avalanche-fuji)"
    )
    asset_address: str = Field(
        description="The address of the token asset"
    )
    amount_required: str = Field(
        description="The required amount, as an uint256 value"
    )
    pay_to_address: str = Field(
        description="The address to pay to. It must be a valid address"
    )
    eip712_domain: dict = Field(
        description="The data related to the eip712 domain for this protocol"
    )



import datetime
from typing import List, Union
from uuid import uuid4
from pydantic import BaseModel
from pymongo import MongoClient, HASHED
from pymongo.collection import Collection
from ..types.payment import SettledPayment
from .base import StorageManager as BaseStorageManager


_BS = 50
_BE = 60


class StorageManager(BaseStorageManager):
    """
    A MongoDB-based storage manager.
    """

    def __init__(self, url: str, database: str, batch_expiration: int = _BE, batch_size: int = _BS):
        super().__init__()
        url = url.strip()
        database = database.strip()
        if not url or not database:
            raise ValueError("Both URL and database must be specified for a"
                             "MongoDB-based StorageManager")
        self._client = MongoClient(url)
        self._database = self._client[database]
        self._batch_expiration = max(_BE, batch_expiration) if isinstance(batch_expiration, int) else _BE
        self._batch_expiration_delta = datetime.timedelta(seconds=self._batch_expiration)
        self._batch_size = max(_BS, batch_size) if isinstance(batch_size, int) else _BS

    def allocate(self, collection: str, payment_id: uuid4,
                 payload: BaseModel, matched_requirements: BaseModel,
                 settled_payment: SettledPayment, webhook_name: str):
        """
        Stores an authorization in 'verified' state. It tells the authorization
        (signed by the `from` sender) and the matched requirement. Ideally, the
        implementation of this method will store the data in some sort of store
        where the matched_requirements are a separate table, and the payment id
        serves as primary key, while the payload contains the whole data.

        These records serve well to track the evolution of a paid requirement
        and also address user claims. This also means: this method must NOT fail
        silently but be violent enough with the payment failing to be stored.

        The storage manager is totally free to choose other fields for the data
        (e.g. some sort of tagging) system.

        Args:
            collection: The collection to store the payment into.
            payment_id: The id of the payment (generated on the fly).
            payload: The client payload fromm headers.
            matched_requirements: The matched requirements.
            settled_payment: The settled payment record. It will be sent
                             via webhook after settlement.
            webhook_name: The associated webhook name to use on launch
                          after settlement.
        """

        collection_: Collection = self._database[collection]

        try:
            collection_.create_index({"payment_id": HASHED}, unique=True)
        except:
            pass

        try:
            collection_.create_index({"webhook_name": HASHED})
        except:
            pass

        try:
            collection_.create_index({"status": HASHED})
        except:
            pass

        collection_.insert_one({
            "payment_id": str(payment_id),
            "payload": payload.model_dump(),
            "matched_requirements": matched_requirements.model_dump(),
            "status": "verified",
            "webhook_payload": settled_payment.model_dump(),
            "webhook_name": webhook_name,
            "created_on": datetime.datetime.now(tz=datetime.UTC)
        })

    def abort(self, collection: str, payment_id: uuid4):
        """
        Aborts a payment id, marking its record.

        Args:
            collection: The collection to remove the payment from.
            payment_id: The id of the payment to remove.
        """

        collection_: Collection = self._database[collection]

        collection_.update_one({"payment_id": str(payment_id)}, {"status": "aborted"})

    def settle(self, collection: str, payment_id: uuid4, transaction: str):
        """
        Confirms a given payment id, meaning that the /settle endpoint worked.

        Args:
            collection: The collection to commit / confirm the payment into.
            payment_id: The id of the payment matching a stored one.
            transaction: The hash of the transaction.
        """

        self._database[collection].update_one(
            {"payment_id": str(payment_id)},
            {"$set": {
                "status": "settled",
                "webhook_payload.transaction_hash": transaction,
                "webhook_payload.settled_on": datetime.datetime.now(tz=datetime.UTC)
            }}
        )

    def _batch_one(self, collection: str, webhook_name: str, worker_id: str,
                   stamp: datetime.datetime) -> bool:
        """
        Tries to find a non-batched record and batches it for the chosen
        worker. The worker will make use of batched records later and then
        proceed to do appropriate payments.

        Args:
            collection: The collection to batch a payment for a worker.
            webhook_name: The name of the webhook the records must have
                          to be batched by this method. Many workers may
                          batch for the same webook name.
            worker_id: The id of the worker to use for batching.
            stamp: The pivot stamp to compute the date.

        Returns:
            Whether one record could be batched or not.
        """

        min_date = stamp - self._batch_expiration_delta
        result = self._database[collection].find_one_and_update({
            "status": "settled",
            "webhook_name": webhook_name,
            "$or": [
                {"batched_on": {"$exists": False}},
                {"batched_on": {"$lte": min_date}},
                {"worker": {"$exists": False}},
                {"worker": {"$in": ["", None]}},
            ]
        }, {"$set": {"worker": worker_id, "batched_on": stamp}})
        return result is not None

    def get_batch(self, collection: str, webhook_name: str, worker_id: str) -> List[SettledPayment]:
        """
        Returns the current batch of records for the given worker and webhook name.

        Args:
            collection: The collection to batch a payment for a worker.
            webhook_name: The name of the webhook the records must have
                          to be batched by this method. Many workers may
                          batch for the same webook name.
            worker_id: The id of the worker to use for batching.

        Returns:
            A list of the requests to send to that webhook.
        """

        # 1. Get the batch size and limit stamps.
        batch_size = self._batch_size
        stamp = datetime.datetime.now(tz=datetime.UTC)
        min_date = stamp - self._batch_expiration_delta

        # 2. Batch remaining items.
        already_batched_count = self._database[collection].count_documents({
            "status": "settled",
            "webhook_name": webhook_name,
            "worker": worker_id,
            "batched_on": {"$gt": min_date}
        })
        if already_batched_count < batch_size:
            for _ in range(batch_size - already_batched_count):
                if not self._batch_one(collection, webhook_name, worker_id, stamp):
                    break

        # 3. Return the records.
        result = []
        cursor = self._database[collection].find({
            "status": "settled",
            "webhook_name": webhook_name,
            "worker": worker_id,
            "batched_on": {"$gt": min_date}
        })
        for document in cursor:
            document.pop("_id", None)
            result.append(SettledPayment(**document["webhook_payload"]))
        return result

    def mark_as_sent(self, collection: str, payment_id: Union[str, uuid4]):
        """
        Marks a payment as webhook-sent.

        Args:
            collection: The collection to mark a payment into.
            payment_id: The payment id.
        """

        self._database[collection].update_one({
            "status": "settled",
            "payment_id": str(payment_id)
        }, {"$set": {"status": "finished"}})



from typing import List, Callable, Awaitable
from flask import Request
from pydantic import Field
from ....core.types.endpoint_settings import X402EndpointSettings as BaseX402EndpointSettings, Y402_ENDPOINT_SETTINGS
from ....core.types.requirements import RequirePaymentDetails


PaymentDetailsListType = List[RequirePaymentDetails] | \
                         Callable[[Request], Awaitable[List[RequirePaymentDetails]] | List[RequirePaymentDetails]]


class X402EndpointSettings(BaseX402EndpointSettings):
    """
    The settings for a single endpoint. It also works as a decorator
    to set the settings into a specific endpoint, which should return
    a quick response based on the reference and nothing else, since
    by its arrival the payment was already sent to the webhook.

    This works on Flask.
    """

    payments_details: PaymentDetailsListType = Field(
        description="Either a non-empty list of allowed payment specs or a callable returning "
                    "a non-empty list of allowed payment specs"
    )

    def __call__(self, endpoint):
        setattr(endpoint, Y402_ENDPOINT_SETTINGS, self)
        return endpoint



from typing import TypedDict


class PaywallConfig(TypedDict, total=False):
    """
    Configuration for paywall UI customization.
    """

    cdp_client_key: str
    app_name: str
    app_logo: str
    session_token_endpoint: str



from typing import List, Union
from uuid import uuid4
from pydantic import BaseModel
from ..types.payment import SettledPayment
from .base import StorageManager as BaseStorageManager


_BS = 50
_BE = 60


class StorageManager(BaseStorageManager):
    """
    A dummy storage manager.
    """

    def allocate(self, collection: str, payment_id: uuid4,
                 payload: BaseModel, matched_requirements: BaseModel,
                 settled_payment: SettledPayment, webhook_name: str):
        """
        Dummy implementation of allocate(collection, payment_id, payload, matched_requirements,
        settled_payment, webhook_name). It's a no-op.
        """

    def abort(self, collection: str, payment_id: uuid4):
        """
        Dummy implementation of abort(collection, payment_id). It's a no-op.
        """

    def settle(self, collection: str, payment_id: uuid4, transaction: str):
        """
        Dummy implementation of settle(collection, payment_id, transaction). It's a no-op.
        """

    def get_batch(self, collection: str, webhook_name: str, worker_id: str) -> List[SettledPayment]:
        """
        Dummy implementation of get_batch(collection, webhook_name, worker_id). It's a no-op.
        """

        return []

    def mark_as_sent(self, collection: str, payment_id: Union[str, uuid4]):
        """
        Dummy implementation of mark_as_sent(collection, payment_id). It's a no-op.
        """



import asyncio
import inspect
import logging
from typing import Any, List
from httpx import AsyncClient
from ..storage.base import StorageManager
from ..types.payment import SettledPayment


LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


async def _maybe_await(result: Any) -> Any:
    if inspect.isawaitable(result):
        result = await result
    return result


async def _send_batch(
    worker_id: str, webhook_name: str, webhook_url: str, api_key: str,
    manager: StorageManager, collection: str, logger: logging.Logger
):
    """
    Performs the processing of a single batch being sent. This updates
    the status of the in-batch processed payments, so new calls to this
    method are not idempotent.

    Args:
        worker_id: The id of this worker.
        webhook_name: The name of the webhook to look records for.
        webhook_url: The URL to send requests to.
        api_key: The api key to use in the X-API-Key header, if any.
        manager: The associated storage manager.
        collection: The collection to use with the manager.
        logger: An optional logger.
    """

    async def _send(client: AsyncClient, payload: SettledPayment):
        try:
            response = await client.post(url=webhook_url, headers=headers, json=payload.model_dump(mode="json"))
            response.raise_for_status()
            await _maybe_await(manager.mark_as_sent(collection, payload.id))
        except:
            logger.exception(f"An exception occurred when processing payment with id={payload.id}")

    api_key = api_key.strip()
    headers = {"X-API-Key": api_key} if api_key else {}
    logger.info("Processing a batch:")
    logger.info("- Retrieving the batch...")
    batch: List[SettledPayment] = await _maybe_await(manager.get_batch(collection, webhook_name, worker_id))
    logger.info(f"- Sending the batch ({len(batch)} elements)...")
    async with AsyncClient(timeout=15) as client:
        await asyncio.gather(*[_send(client, payload) for payload in batch])
    logger.info("- Updating the batch...")
    for element in batch:
        manager.mark_as_sent(collection, element.id)


async def _webhook_worker(
    worker_id: str, webhook_name: str, webhook_url: str, api_key: str,
    manager: StorageManager, collection: str, logger: logging.Logger = None,
    sleep_time: int = 5
):
    """
    Sends a batch of settled requests, assigned to a webhook name, to the
    corresponding webhook URL (and with proper authentication, if any).
    It uses a specific worker name for this, so it does not collide with
    other workers in the same process.

    Args:
        worker_id: The id of this worker.
        webhook_name: The name of the webhook to look records for.
        webhook_url: The URL to send requests to.
        api_key: The api key to use in the X-API-Key header, if any.
        manager: The associated storage manager.
        collection: The collection to use with the manager.
        logger: An optional logger.
        sleep_time: A sleep time between iterations.
    """

    logger = logger or LOGGER
    
    logger.info(f"Starting webhook worker loop.\n- webhook_name={webhook_name}\n- worker_id={worker_id}")
    try:
        while True:
            await _send_batch(worker_id, webhook_name, webhook_url, api_key, manager, collection, logger)
            await asyncio.sleep(sleep_time)
    except KeyboardInterrupt:
        logger.info(f"Worker loop cancelled by the user")
    except:
        logger.exception("Worker loop stopped because of an error:")


def webhook_worker(
    worker_id: str, webhook_name: str, webhook_url: str, api_key: str,
    storage_manager: StorageManager, collection: str, logger: logging.Logger = None,
    sleep_time: int = 5
):
    """
    Sends a batch of settled requests, assigned to a webhook name, to the
    corresponding webhook URL (and with proper authentication, if any).
    It uses a specific worker name for this, so it does not collide with
    other workers in the same process.

    Args:
        worker_id: The id of this worker.
        webhook_name: The name of the webhook to look records for.
        webhook_url: The URL to send requests to.
        api_key: The api key to use in the X-API-Key header, if any.
        storage_manager: The associated storage manager.
        collection: The collection to use with the manager.
        logger: An optional logger.
        sleep_time: A sleep time between iterations.
    """

    asyncio.run(_webhook_worker(worker_id, webhook_name, webhook_url, api_key, storage_manager, collection,
                                logger, sleep_time))



import base64
import functools
import traceback
from typing import Optional, List, Literal
import logging
from flask import request, g, make_response
from pydantic import validate_call
from werkzeug.exceptions import HTTPException
from .prices import compute_prices
from .request_data import get_root_url
from .response_presets import x402_response, response
from .types.endpoint_settings import X402EndpointSettings, Y402_ENDPOINT_SETTINGS
from ...storage.base import StorageManager
from ...storage.dummy import StorageManager as DummyStorageManager
from ...core.types.facilitator import FacilitatorConfig
from ...core.types.paywall import PaywallConfig
from ...core.types.registry import FinalEndpointSetupRegistry
from ...core.types.requirements import FinalRequiredPaymentDetails, PaymentRequirements
from ...core.types.setup import Y402Setup
from ...core.utils.headers import decode_payment_header, validate_payment_asset
from ...core.utils.prices import PriceComputingError


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _endpoint_invoker(endpoint, reference, custom_paywall_html_, paywall_config_,
                      chain_id_by_name, args, kwargs):
    def f(payment_id):
        # 1. As state, keep: The payment_id, the send payment error (if any),
        #    and the reference (it might be blank).
        g.x402 = {
            "payment_id": payment_id,
            "reference": reference
        }

        # 2. Call and wrap the underlying endpoint, which should have a very small logic.
        try:
            response_ = make_response(endpoint(*args, **kwargs))
        except Exception:
            logger.exception(traceback.format_exc())
            response_ = response(
                500,
                "An error occurred, but a payment was already processed. Contact support "
                "to claim your product or service by the internal payment "
                f"id: {payment_id}", custom_paywall_html_, paywall_config_, [], chain_id_by_name
            )

        return response_, 200 <= response_.status_code < 400
    return f


@validate_call
def payment_required(
    *,
    mime_type: str = "",
    default_max_deadline_seconds: int = 60,
    paywall_config: Optional[PaywallConfig] = None,
    custom_paywall_html: Optional[str] = None,
    facilitator_config: Optional[FacilitatorConfig] = None,
    setup: Optional[Y402Setup] = None,
    client_http_library: Literal["httpx_sync", "requests"] = "httpx_sync",
    storage_manager: Optional[StorageManager] = None
):
    """
    This factory creates a middleware that performs the
    full Y402 (extremely similar and compatible to X402)
    cycle, directly to be used in a FastAPI application.

    Args:
        mime_type: The MIME type for the answer.
        default_max_deadline_seconds: The maximum time the client can take to send the paid request.
        paywall_config: The paywall configuration (applies when serving HTML).
        custom_paywall_html: The custom paywall HTML contents (applies when serving HTML).
        facilitator_config: The configuration for the facilitator.
        setup: The general, middleware-wide, setup.
        client_http_library: The allowed client library (to make new HTTP calls with).
        storage_manager: The default storage manager. If not set, it will use a Dummy storage
                         manager, which is a no-op. Non-dummy Storage managers are used to store
                         the requested and paid-for jobs. Using a dummy storage manager forfeits
                         the webhook-related logic and relies on standard x402 practices.

    Returns:
        A middleware function, to be used in an existing Flask app (and returning
        a new app).
    """

    storage_manager = storage_manager or DummyStorageManager()
    registry = FinalEndpointSetupRegistry(setup)

    def decorator(endpoint):
        """
        This decorator wraps an endpoint function into an endpoint that has a previous
        setup of x402 crypto payment processing. Typically, you'll instantiate this
        decorator once, like:

            decorator = payment_required(...)

        And then apply it in all of your code. For each endpoint, apply this decorator
        immediately above @X402EndpointSettings(...) and somewhere below any @app.route-like
        decorator. As an example, do this:

            @app.route(...)
            ...
            @decorator
            @X402EndpointSettings(...)
            def my_endpoint(...):
                ...

        Args:
            endpoint: The endpoint function to wrap.

        Returns:
            The wrapped endpoint.
        """

        @functools.wraps(endpoint)
        def new_endpoint(*args, **kwargs):
            """
            The new endpoint function. Forwards all the endpoint arguments, and
            contains all the related payment logic prior to performing any call
            to the original endpoint.
            """

            # 1. Retrieve the endpoint, and the endpoint data.
            endpoint_data: Optional[X402EndpointSettings] = getattr(endpoint, Y402_ENDPOINT_SETTINGS, None)
            if not isinstance(endpoint_data, X402EndpointSettings):
                logger.warning("The endpoint to apply this `payment_required(...)` decorator must be one "
                               "decorated with a @X402EndpointSettings(...) instance. This endpoint has "
                               "no @X402EndpointSettings(...) decorator, so no x402 payment logic will "
                               f"be applied to this endpoint (url is: {request.url})")
                return endpoint(*args, **kwargs)

            # 2. Get / initialize the per-endpoint aggregated networks
            #    and tokens setup data, and the resource metadata.
            resource_url = get_root_url() + "/" + endpoint_data.resource_url.lstrip("/") \
                if endpoint_data.resource_url else str(request.url)
            description = endpoint_data.description or ""
            max_deadline_seconds = endpoint_data.max_deadline_seconds or default_max_deadline_seconds
            output_schema = endpoint_data.output_schema
            input_schema = endpoint_data.input_schema
            mime_type_ = endpoint_data.mime_type or mime_type or "application/json"
            paywall_config_ = endpoint_data.paywall_config or paywall_config
            custom_paywall_html_ = endpoint_data.custom_paywall_html or custom_paywall_html
            storage_collection = endpoint_data.storage_collection.strip() or "payments"

            if storage_manager is None:
                logger.error(f"Storage manager not defined")
                return response(
                    500, f"The resource {resource_url} is not properly configured",
                    custom_paywall_html_, paywall_config_, [], {}
                )

            reference = ''
            if endpoint_data.reference_param is not None:
                reference = request.view_args.get(endpoint_data.reference_param, None)
                if reference is None:
                    logger.error(f"Path parameter '{endpoint_data.reference_param}' not defined for "
                                 f"resource {resource_url}")
                    return response(
                        500, f"The resource {resource_url} is not properly configured",
                        custom_paywall_html_, paywall_config_, [], {}
                    )

            # 3. Given the per-endpoint configuration, get all the prices
            #    the user can pay. All of them will be *exact*, but can
            #    express in different networks (existing; configured ones).
            #    This said, each record will have:
            #    - The network (it's name, not the chain id; unique).
            #    - The address of the token.
            #    - The required amount (we'll use "exact" here).
            #    - The pay-to address.
            try:
                merged_setup: Y402Setup = registry[endpoint]
                chain_id_by_name = merged_setup.get_chain_ids_mapping()
                prices: List[FinalRequiredPaymentDetails] = compute_prices(
                    request, endpoint_data.payments_details, merged_setup
                )
            except HTTPException:
                logger.exception("An early-terminating HTTP exception occurred at price computing stage:")
                raise
            except PriceComputingError as e:
                logger.exception("An error occurred at price computing stage:")
                return response(
                    500, f"The resource {resource_url}'s setup / pricing is not properly configured: {str(e)}",
                    custom_paywall_html_, paywall_config_, [], {}
                )
            except:
                logger.exception("An error occurred at price computing stage:")
                return response(
                    500, f"The resource {resource_url}'s setup / pricing is not properly configured",
                    custom_paywall_html_, paywall_config_, [], {}
                )

            # 4. Construct payment details. Only one payment is supported
            # per network, since even when it's not a problem in the protocol
            # itself, most of the clients will stick to this library implementation.
            payment_requirements = [
                PaymentRequirements(
                    scheme=price.scheme,  # e.g. "exact"
                    network=price.network,  # e.g. "ethereum"
                    asset=price.asset_address,  # 0x...40 hex digits...
                    max_amount_required=price.amount_required,  # ...dec digits...
                    resource=resource_url,
                    description=description,
                    mime_type=mime_type_,
                    pay_to=price.pay_to_address,  # 0x...40 hex digits...
                    max_timeout_seconds=max_deadline_seconds,
                    output_schema={
                        "input": {
                            "type": "http",
                            "method": request.method.upper(),
                            "discoverable": True,
                            **(input_schema.model_dump() if input_schema else {}),
                        },
                        "output": output_schema,
                    },
                    extra=price.eip712_domain,
                ) for price in prices
            ]

            # 5. Check for payment header. If it does not exist, halt and require
            #    a payment.
            payment_header = request.headers.get("X-PAYMENT", "")
            if payment_header == "":
                return x402_response("No X-PAYMENT header provided", custom_paywall_html_,
                                     paywall_config_, payment_requirements, chain_id_by_name)

            # 6. Extract the payment header.
            try:
                payment = decode_payment_header(payment_header)
            except Exception:
                logger.exception(
                    f"Invalid payment header format from {request.remote_addr}:"
                )
                return x402_response("Invalid payment header format", custom_paywall_html_,
                                     paywall_config_, payment_requirements, chain_id_by_name)

            # 7. Extract the extra header, perhaps, with payment token. It
            #    must be an address, if present.
            payment_asset_header = request.headers.get("X-PAYMENT-ASSET", "").lower()

            # 8. Based on the payment_token_header, if available, select
            #    which token was selected. Otherwise, iterate until a
            #    token matches the signature.
            network = payment.network
            code, asset, ok = validate_payment_asset(network, payment, payment_asset_header, merged_setup)
            if not ok:
                logger.error(
                    f"Invalid payment header format from {request.remote_addr}:"
                )
                return x402_response("Invalid payment asset", custom_paywall_html_,
                                     paywall_config_, payment_requirements, chain_id_by_name)

            # 9. Pick the proper payment by the code, and make use of it later.
            requirement = next(
                (requirement
                 for requirement in payment_requirements
                 if requirement.asset == asset and requirement.network == network),
                None
            )
            if not requirement:
                return x402_response("Invalid payment asset", custom_paywall_html_,
                                     paywall_config_, payment_requirements, chain_id_by_name)

            # 10. Pick the proper payment processor adapter.
            match client_http_library:
                case "httpx_sync":
                    from ...lifecycle.httpx_sync import process_payment
                case "requests":
                    from ...lifecycle.requests import process_payment
                case _:
                    return x402_response("Server not properly configured", custom_paywall_html_,
                                         paywall_config_, payment_requirements, chain_id_by_name)

            # 11. Actually process the payment.
            try:
                invoke_endpoint = _endpoint_invoker(
                    endpoint, reference, custom_paywall_html_, paywall_config_,
                    chain_id_by_name, args, kwargs
                )
                payment_id, settle_response, response_ = process_payment(
                    resource_url, endpoint_data.tags, reference, invoke_endpoint, payment,
                    requirement, merged_setup, facilitator_config,
                    storage_manager, storage_collection, endpoint_data.webhook_name
                )
                if settle_response and not settle_response.success:
                    return x402_response("Settle failed: " + (settle_response.error_reason or "Unknown error"),
                                         custom_paywall_html_, paywall_config_, payment_requirements,
                                         chain_id_by_name)
            except:
                logger.exception("An exception occurred when interacting with the facilitator or forwarding "
                                 "the payment:")
                return x402_response("The payment was invalid or there was an error processing it",
                                     custom_paywall_html_, paywall_config_, payment_requirements, chain_id_by_name)

            # 12. Finally, return the internal response.
            if 200 <= response_.status_code < 400:
                response_.headers["X-PAYMENT-RESPONSE"] = base64.b64encode(
                    settle_response.model_dump_json(by_alias=True).encode("utf-8")
                ).decode("utf-8")
            return response_

        return new_endpoint

    return decorator



KNOWN_NETWORKS_AND_TOKENS = {
    "base": {
        "chain_id": 8453,
        "tokens": {
            "usdc": {
                "address": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0x60a3E35Cc302bFA44Cb288Bc5a4F316Fdb1adb42",
                "name": "EURC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "base-sepolia": {
        "chain_id": 84532,
        "tokens": {
            "usdc": {
                "address": "0x036CbD53842c5426634e7929541eC2318f3dCF7e",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0x808456652fdb597867f38412077A9182bf77359F",
                "name": "EURC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "avalanche": {
        "chain_id": 43114,
        "tokens": {
            "usdc": {
                "address": "0xB97EF9Ef8734C71904D8002F8b6Bc66Dd9c48a6E",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0xc891eb4cbdeff6e073e859e987815ed1505c2acd",
                "name": "Euro Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "avalanche-fuji": {
        "chain_id": 43113,
        "tokens": {
            "usdc": {
                "address": "0x5425890298aed601595a70AB815c96711a31Bc65",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0x5E44db7996c682E92a960b65AC713a54AD815c6B",
                "name": "EUR Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "sei": {
        "chain_id": 1329,
        "tokens": {
            "usdc": {
                "address": "0xe15fC38F6D8c56aF07bbCBe3BAf5708A2Bf42392",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    "sei-testnet": {
        "chain_id": 1328,
        "tokens": {
            "usdc": {
                "address": "0x4fCF1784B31630811181f670Aea7A7bEF803eaED",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    "polygon": {
        "chain_id": 137,
        "tokens": {
            "usdc": {
                "address": "0x3c499c542cEF5E3811e1192ce70d8cC03d5c3359",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    "polygon-amoy": {
        "chain_id": 80002,
        "tokens": {
            "usdc": {
                "address": "0x41E94Eb019C0762f9Bfcf9Fb1E58725BfB0e7582",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    # These are local development networks only.
    "local": {
        "chain_id": 31337,
        "tokens": {}
    }
}



from typing import Optional, Dict, List
from httpx import Request, Response, Client
from eth_account import Account
from .common import (
    Y402Client as BaseY402Client,
    MissingRequestConfigError,
    PaymentError,
    PaymentSelectorCallable, decode_x_payment_networks,
)
from ..core.types.responses import x402PaymentRequiredResponse


class HttpxHooks:
    """
    Hooks for the Httpx sync adapter.
    """

    def __init__(self, client: BaseY402Client):
        self.client = client
        self._is_retry = False

    def on_request(self, request: Request):
        """
        Handle request before it is sent.
        """

    def on_response(self, response: Response) -> Response:
        """
        Handle response after it is received.
        """

        # If this is not a 402, just return the response.
        if response.status_code != 402:
            return response

        # If this is a retry response, just return it.
        if self._is_retry:
            return response

        try:
            if not response.request:
                raise MissingRequestConfigError("Missing request configuration")

            # Read the response content before parsing.
            response.read()

            data = response.json()

            # Get the X-Payment-Networks response.
            x_payment_networks = response.headers.get("X-Payment-Networks")
            try:
                y402_chain_id_by_name = decode_x_payment_networks(x_payment_networks) if x_payment_networks else None
            except:
                y402_chain_id_by_name = None

            payment_response = x402PaymentRequiredResponse(**data)

            # Select payment requirements.
            selected_requirements = self.client.select_payment_requirements(
                payment_response.accepts
            )

            # Create payment header.
            payment_header = self.client.create_payment_header(
                selected_requirements, payment_response.x402_version, y402_chain_id_by_name
            )

            # Mark as retry and add payment header.
            self._is_retry = True
            request = response.request

            request.headers["X-Payment"] = payment_header
            request.headers["Access-Control-Expose-Headers"] = "X-Payment-Response, X-Payment-Networks"

            # Retry the request.
            with Client() as client:
                retry_response = client.send(request)

                # Copy the retry response data to the original response
                response.status_code = retry_response.status_code
                response.headers = retry_response.headers
                response._content = retry_response._content
                return response
        except PaymentError as e:
            self._is_retry = False
            raise e
        except Exception as e:
            self._is_retry = False
            raise PaymentError(f"Failed to handle payment: {str(e)}") from e


def y402_payment_hooks(
    account: Account,
    payment_requirements_selector: Optional[PaymentSelectorCallable] = None,
    chain_id_by_name: Optional[Dict[str, int]] = None
) -> Dict[str, List]:
    """Create httpx event hooks dictionary for handling 402 Payment Required responses.

    Args:
        account: eth_account.Account instance for signing payments.
        payment_requirements_selector: Optional custom selector for payment requirements.
            Should be a callable that takes (accepts, network_filter, scheme_filter, max_value)
            and returns a PaymentRequirements object.
        chain_id_by_name: An optional mapping of extra networks to map (useful when there
                          are more networks to be used by the server). This is useful for
                          when the remote x402 server implements custom networks and does
                          not implement y402, for y402 implementations will tell another
                          way to establish that mapping to y402-compatible clients (like
                          this one).

    Returns:
        Dictionary of event hooks that can be directly assigned to client.event_hooks.
    """

    # Create the Y402 client.
    client = BaseY402Client(
        account,
        payment_requirements_selector=payment_requirements_selector,
        chain_id_by_name=chain_id_by_name
    )

    # Create hooks.
    hooks = HttpxHooks(client)

    # Return event hooks dictionary.
    return {
        "request": [hooks.on_request],
        "response": [hooks.on_response],
    }


class Y402Client(Client):
    """
    Client with built-in y402 payment handling.
    """

    def __init__(
        self,
        account: Account,
        payment_requirements_selector: Optional[PaymentSelectorCallable] = None,
        chain_id_by_name: Optional[Dict[str, int]] = None,
        **kwargs
    ):
        """
        Initialize a Client with x402 payment handling (plus y402 enhancements).

        Args:
            account: eth_account.Account instance for signing payments.
            payment_requirements_selector: Optional custom selector for payment requirements.
                Should be a callable that takes (accepts, network_filter, scheme_filter)
                and returns a PaymentRequirements object.
            chain_id_by_name: An optional mapping of extra networks to map (useful when there
                              are more networks to be used by the server). This is useful for
                              when the remote x402 server implements custom networks and does
                              not implement y402, for y402 implementations will tell another
                              way to establish that mapping to y402-compatible clients (like
                              this one).
            **kwargs: Additional arguments to pass to Client.
        """

        super().__init__(**kwargs)
        self.event_hooks = y402_payment_hooks(
            account, payment_requirements_selector, chain_id_by_name
        )



from .base import FacilitatorClient as BaseFacilitatorClient
from ..core.types.errors import ConditionalDependencyError, BaseError
from .errors import VerifyFacilitatorUnknownError, SettleFacilitatorUnknownError
from ..core.types.facilitator import VerifyRequest, VerifyResponse, SettleResponse, SettleRequest


try:
    import httpx
except ImportError:
    raise ConditionalDependencyError("httpx library is not installed. Install it as a requirement "
                                     "by invoking httpx==0.28.1 or similar")


class FacilitatorClient(BaseFacilitatorClient):
    """
    This class stands for a httpx-based facilitator client.
    The results come in the form of awaitable objects.
    """

    async def verify(self, request: VerifyRequest, timeout: int = 10) -> VerifyResponse:
        """
        Performs a /verify POST call with the given data.

        Args:
            request: The current request.
            timeout: The timeout.

        Returns:
            The verify response.
        """

        headers = self._make_headers('verify')
        if timeout < 1:
            timeout = 1
        try:
            async with httpx.AsyncClient(timeout=15) as client:
                response = await client.post(self._config.url.rstrip("/") + "/verify", headers=headers,
                                             json=request.to_json(), timeout=timeout)
                response.read()
                self._check_verify_status(response.status_code, response.content, response.headers.get('Content-Type'))
                return self._parse_verify_obj(response.json())
        except BaseError:
            raise
        except Exception as e:
            raise VerifyFacilitatorUnknownError(e)

    async def settle(self, request: SettleRequest, timeout: int = 10) -> SettleResponse:
        """
        Performs a /settle POST call with the given data.

        Args:
            request: The current request.
            timeout: The timeout.

        Returns:
            The settle response.
        """

        headers = self._make_headers('settle')
        if timeout < 1:
            timeout = 1
        try:
            async with httpx.AsyncClient(timeout=15) as client:
                response = await client.post(self._config.url.rstrip("/") + "/settle", headers=headers,
                                             json=request.to_json(), timeout=timeout)
                response.read()
                self._check_settle_status(response.status_code, response.content, response.headers.get('Content-Type'))
                return self._parse_settle_obj(response.json())
        except BaseError:
            raise
        except Exception as e:
            raise SettleFacilitatorUnknownError(e)



import inspect
from typing import List, Callable
from fastapi import Request
from ...core.types.requirements import RequirePaymentDetails, FinalRequiredPaymentDetails
from ...core.types.setup import Y402Setup
from ...core.utils.prices import resolve_final_payment


async def compute_prices(
    request: Request,
    prices: List[RequirePaymentDetails] | Callable[[Request], List[RequirePaymentDetails]],
    setup: Y402Setup
) -> List[FinalRequiredPaymentDetails]:
    """
    Args:
        request: The current request.
        prices: Either the prices or a callable based on the request and returning the prices.
                Always one single price per network.
        setup: The underlying, final, Y402Setup to compute the prices from.

    Returns:
        A list of final required payment details.
    """

    if callable(prices):
        prices = prices(request)
        if inspect.isawaitable(prices):
            prices = await prices

    return [resolve_final_payment(price, setup) for price in prices]



from typing import Dict, Literal, Callable, Awaitable, Optional
from pydantic import BaseModel, Field, ConfigDict
from pydantic.alias_generators import to_camel
from .constants import X402_VERSION
from .client import PaymentPayload
from .requirements import PaymentRequirements


class BaseRequest(BaseModel):
    """
    This class stands for a base request for the facilitator endpoints.
    """

    x402_version: int = Field(
        default=X402_VERSION, alias="x402Version",
        description="The involved x402 version, for compatibility with the protocol"
    )
    payment_payload: PaymentPayload = Field(
        alias="paymentPayload",
        description="The client-sent payload"
    )
    payment_requirements: PaymentRequirements = Field(
        alias="paymentRequirements",
        description="The allowed requirements for this payment"
    )

    def to_json(self):
        """
        Builds the JSON payload of this base request object.

        Returns:
            A dictionary with the JSON response.
        """

        return {
            "x402Version": self.x402_version,
            "paymentPayload": self.payment_payload.model_dump(by_alias=True),
            "paymentRequirements": self.payment_requirements.model_dump(
                by_alias=True, exclude_none=True
            ),
        }


class VerifyRequest(BaseRequest):
    """
    This class stands for the body a request to the /verify endpoint.
    """


class VerifyResponse(BaseModel):
    """
    This class stands for the body of a /verify response.
    """

    is_valid: bool = Field(alias="isValid")
    invalid_reason: Optional[str] = Field(default=None, alias="invalidReason")
    payer: Optional[str] = Field(default=None)

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )


class SettleRequest(BaseRequest):
    """
    This class stands for the body of a request to the /settle endpoint.
    """


class SettleResponse(BaseModel):
    """
    This class stands for the body of a /settle response.
    """

    success: bool
    error_reason: Optional[str] = Field(None, alias="errorReason")
    transaction: Optional[str] = None
    network: Optional[str] = None
    payer: Optional[str] = None

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )


FacilitatorHeaders = Dict[Literal['settle', 'verify'], dict | Callable[[], dict] | Callable[[], Awaitable[dict]]]
FacilitatorHeaders.__doc__ = """
The way the headers are built when performing requests to
an x402 facilitator. They can be per endpoint and of types:

1. A dictionary.
2. A callable returning a dictionary of headers.
3. An awaitable callable returning a dictionary of headers.
"""


class FacilitatorConfig(BaseModel):
    """
    Configuration for the X402 facilitator service.

    Attributes:
        url: The base URL for the facilitator service.
        headers: Optional function to create authentication headers.
    """

    url: str = Field(
        default="https://x402.org/facilitator",
        description="The URL of the facilitator. It does not need to be a top-level URL but must "
                    "have two facilitator-compatible sub-endpoints named /verify and /settle"
    )
    headers: Optional[FacilitatorHeaders] = Field(
        default=None,
        description="The headers, or header-generation callables, to generate the headers for "
                    "each of the endpoints"
    )



from typing import List, Union
from uuid import uuid4
from pydantic import BaseModel
from ..types.payment import SettledPayment


class StorageManager(BaseModel):
    """
    This class defines a storage manager for the (post-verified) user payment requests.
    """

    def allocate(self, collection: str, payment_id: uuid4,
                 payload: BaseModel, matched_requirements: BaseModel,
                 settled_payment: SettledPayment, webhook_name: str):
        """
        Stores an authorization in 'verified' state. It tells the authorization
        (signed by the `from` sender) and the matched requirement. Ideally, the
        implementation of this method will store the data in some sort of store
        where the matched_requirements are a separate table, and the payment id
        serves as primary key, while the payload contains the whole data.

        These records serve well to track the evolution of a paid requirement
        and also address user claims. This also means: this method must NOT fail
        silently but be violent enough with the payment failing to be stored.

        The storage manager is totally free to choose other fields for the data
        (e.g. some sort of tagging) system.

        Args:
            collection: The collection to store the payment into.
            payment_id: The id of the payment (generated on the fly).
            payload: The client payload fromm headers.
            matched_requirements: The matched requirements.
            settled_payment: The settled payment record. It will be sent
                             via webhook after settlement.
            webhook_name: The associated webhook name to use on launch
                          after settlement.
        """

        raise NotImplementedError

    def abort(self, collection: str, payment_id: uuid4):
        """
        Aborts a payment id, marking its record.

        Args:
            collection: The collection to remove the payment from.
            payment_id: The id of the payment to mark.
        """

        raise NotImplementedError

    def settle(self, collection: str, payment_id: uuid4, transaction: str):
        """
        Confirms a given payment id, meaning that the /settle endpoint worked.

        Args:
            collection: The collection to settle the payment into.
            payment_id: The id of the payment matching a stored one.
            transaction: The hash of the transaction.
        """

        raise NotImplementedError

    def get_batch(self, collection: str, webhook_name: str, worker_id: str) -> List[SettledPayment]:
        """
        Returns the current batch of records for the given worker and webhook name.

        Args:
            collection: The collection to batch a payment for a worker.
            webhook_name: The name of the webhook the records must have
                          to be batched by this method. Many workers may
                          batch for the same webook name.
            worker_id: The id of the worker to use for batching.

        Returns:
            A list of the requests to send to that webhook.
        """

        raise NotImplementedError

    def mark_as_sent(self, collection: str, payment_id: Union[str, uuid4]):
        """
        Marks a payment as webhook-sent.

        Args:
            collection: The collection to mark a payment into.
            payment_id: The payment id.
        """

        raise NotImplementedError



import traceback
from ..types.errors import MisconfigurationError
from ..types.requirements import FinalRequiredPaymentDetails, RequirePaymentDetails, Price, TokenAmount
from ..types.setup import Y402Setup


class PriceComputingError(Exception):
    """
    Tells an error occurred while computing the price
    of an endpoint.
    """


def _resolve_payment_price(
    network: str, price: Price, setup: Y402Setup
):
    if isinstance(price, str):
        try:
            code, amount = setup.parse_price_label(network, price)
            name, _, address, version, _ = setup.get_token_metadata(network, code)
            return amount, address, {"name": name, "version": version}
        except Exception:
            traceback.print_exc()
            raise PriceComputingError("There was an error while computing a price from string")
    elif isinstance(price, int):
        code = setup.get_default_token(network)
        if code is None:
            raise MisconfigurationError(f"The network {network} does not have a default token")
        try:
            name, _, address, version, _ = setup.get_token_metadata(network, code)
        except Exception:
            traceback.print_exc()
            raise PriceComputingError("There was an error while computing a price from int")
        return str(price), address, {"name": name, "version": version}
    elif isinstance(price, TokenAmount):
        # TokenAmount type - already in atomic units with asset info.
        return (
            price.amount,
            price.asset.address,
            {
                "name": price.asset.eip712.name,
                "version": price.asset.eip712.version,
            },
        )
    else:
        raise ValueError(f"Invalid price type: {type(price)}")


def resolve_final_payment(
    required_payment: RequirePaymentDetails,
    setup: Y402Setup
) -> FinalRequiredPaymentDetails:
    """
    Resolves a final payment based on a price specification.

    Args:
        required_payment: The required payment to base the final payment on.
        setup: The final setup for an endpoint.

    Returns:
        The final payment requirement.
    """

    amount, address, eip712_domain = _resolve_payment_price(
        required_payment.network, required_payment.price, setup
    )
    return FinalRequiredPaymentDetails(
        scheme=required_payment.scheme,
        network=required_payment.network,
        pay_to_address=required_payment.pay_to_address,
        asset_address=address,
        amount_required=amount,
        eip712_domain=eip712_domain
    )


from datetime import datetime
from typing import List, Optional
from pydantic import Field, BaseModel


class PaymentIdentity(BaseModel):
    """
    This class holds the identity of a payment.
    """

    resource: str = Field(
        description="The URL of the resource. This URL is both user-defined "
                    "and user-configured. It has exactly one argument named "
                    "reference_ticket which can refer an object that can be "
                    "paid once or more (depending on the asset)"
    )
    tags: List[str] = Field(
        description="This is a mandatory list of tags which can be used to "
                    "identify categories for the payments. Tags are set once "
                    "per configuration and are static in nature, since they "
                    "are categories of payment processing. By configuration, "
                    "one resource URL will relate to a set of tags and a given "
                    "(internal) reference"
    )
    reference: str = Field(
        description="This is the internal reference code, primarily added "
                    "into the resource URL, as it is used to identify the "
                    "payment"
    )


class PaymentDetails(BaseModel):
    """
    This class stands for the payment details (source address, amount, token, ...).
    """

    # Canonic fields.
    payer: str = Field(description="The 0x-prefixed ethereum address of the account authorizing the payment")
    chain_id: str = Field(description="The id of the chain, in decimal representation")
    token: str = Field(description="The 0x-prefixed ethereum address of the token contract. It will support "
                                   "ERC-3009 or be somehow supported by this payment system")
    value: str = Field(description="The decimal representation of the token amount. It will be expressed in "
                                   "the minimal token units according to the decimals supported by the token")
    pay_to_address: str = Field(description="The address that received the payment")

    # Representational fields.
    code: str = Field(description="The codename of the token (e.g. usdt, usdc, eurc, eurt)")
    name: str = Field(description="The display name of the token (e.g. USDt, USDC, EURt, EURC")
    price_label: str = Field(
        description="A symbol-based representation of the amount being paid. For example, USDC and USDt "
                    "will represent 150000 as $0.15, while EURC and and EURt will represent 150000 as "
                    "€0.15. Tokens without symbol will use no symbol, representing a fractional number "
                    "according to their decimals, e.g. a 10-decimals token will represent 150000 as "
                    "0.000015, without any kind of symbol unless one is stated"
    )


class SettledPayment(BaseModel):
    """
    This class stands for a settled payment. Settled
    payments are notified via a custom webhook URL
    (POST method) so the customers can make use of
    it when attending the request.
    """

    id: str = Field(
        description="The unique ID of the payment"
    )
    version: int = Field(
        description="The y402 version"
    )
    identity: PaymentIdentity = Field(
        description="The identity of the payment (with identifier and category)"
    )
    details: PaymentDetails = Field(
        description="The details of the payment (token, network, ...)"
    )
    settled_on: Optional[datetime] = Field(
        default=None,
        description="The time this payment was settled on"
    )
    transaction_hash: Optional[str] = Field(
        default=None,
        description="The 0x-prefixed transaction hash"
    )
